/* --------------------------------------------------------------
    Integral arithmetical calculator.

    http://www.imach.uran.ru/cbignum

    Copyright 2002-2005 by Raul N.Shakirov, IMach of RAS(UB).
    All Rights Reserved.

    Permission has been granted to copy, distribute and modify
    software in any context without fee, including a commercial
    application, provided that the aforesaid copyright statement
    is present here as well as exhaustive description of changes.

    THE SOFTWARE IS DISTRIBUTED "AS IS". NO WARRANTY OF ANY KIND
    IS EXPRESSED OR IMPLIED. YOU USE AT YOUR OWN RISK. THE AUTHOR
    WILL NOT BE LIABLE FOR DATA LOSS, DAMAGES, LOSS OF PROFITS OR
    ANY OTHER KIND OF LOSS WHILE USING OR MISUSING THIS SOFTWARE.
-------------------------------------------------------------- */
#ifdef  _MSC_VER
#pragma warning(disable:4995)
#endif/*_MSC_VER*/
#include <assert.h>
#include <ctype.h>
#include <fstream.h>
#include <iomanip.h>
#include <stdlib.h>
#include <string.h>
#include "Cbignum.h"
#include "Gettimer.h"
#include "Random3.h"

static  int GenerateRandom (cBigNumber& a, size_t radix, size_t maxexp);

const char szID[] =
    "Big integer calculator. Version 1.1a beta public of "__DATE__"\n"
    "Copyright 2002-2005 by Dr. Raul N.Shakirov, IMach of RAS(UB)\n"
    ;
const char szHelp[] =
    "Call:  Arifexp [Options]\n"
    "Options:\n"
    "-idle  idle priority\n"
    "-high  idle priority\n"
    "-hex   hexadecimal input/output\n"
    "-hexi  hexadecimal input\n"
    "-hexo  hexadecimal output\n"
    "-check check results of binary operations, if possible\n"
    "-rand  generate random numbers instead of computing:\n"
    "       positive number of length n instead of value n\n"
    "       negative number of length n instead of value -n\n"
    "-exp n append up to n trailing zeroes to random numbers\n"
    "-help  show this help and exit\n"
    "\n"
    "Site:  http://www.imach.uran.ru/cbignum\n"
    ;

int main (int argc, char** argv)
{
  int radix = 10;
  int maxexp = 0;
  cin >> dec;
  cout << dec;

  int retcode = 0;
  long total1 = 0;
  long total2 = 0;
  long total3 = 0;
  long total4 = 0;
  long total5 = 0;

  cerr << szID << endl;

  // Look for arguments.

  char *pszPriority = "";
  char *pszGenerate = "";
  char *pszCheck    = "";

  for (; argc > 1; --argc, ++argv)
  {
    if (strcmp (argv [1], "-idle") == 0)
    {
      pszPriority = "Idle priority\n\n";
      SetIdlePriority();
      continue;
    }
    if (strcmp (argv [1], "-high") == 0)
    {
      pszPriority = "High priority\n\n";
      SetHighPriority();
      continue;
    }
    if (strcmp (argv [1], "-hex") == 0)
    {
      cin >> hex;
      cout << hex; radix = 16;
      continue;
    }
    if (strcmp (argv [1], "-hexi") == 0)
    {
      cin >> hex;
      continue;
    }
    if (strcmp (argv [1], "-hexo") == 0)
    {
      cout << hex; radix = 16;
      continue;
    }
    if (strcmp (argv [1], "-check") == 0)
    {
      pszCheck = "Check mode\n\n";
      continue;
    }
    if (strcmp (argv [1], "-rand") == 0)
    {
      RRand3();
      pszGenerate = "Generating random numbers\n\n";
      continue;
    }
    if (strcmp (argv [1], "-exp") == 0 && argc > 2)
    {
      maxexp = atoi (argv [2]);
      --argc, ++argv;
      continue;
    }
    if (strcmp (argv [1],  "-help") == 0 ||
        strcmp (argv [1],  "-h")    == 0 ||
        strcmp (argv [1],  "-?")    == 0)
    {
      cout << szHelp;
      return 0;
    }

    // Unrecognized option.

    cerr << "Unrecognized option " << argv [1] << endl;
    return 1;
  }

  cerr << pszPriority;
  cerr << pszGenerate;

  while (cin)
  {
    cerr << "Please, type expression "
         << "a=, a+-*/%\\|&^b=, a+b*\\c=, a*\\b%c=, a>>b, a<<b or 2Va=\n"
         << "and press Enter or press Enter to exit\n";

    long timer0 = GetTimer (0);

    char ch = 0; cin.get (ch);
    if (cin.eof() || ch == '\n') break;
    cin.putback (ch);
    ch = 0;

    cBigNumber a = 0, b = 0, c = 0; // Operands
    int op = 0, md = 0;             // Operators
    unsigned error = 0;             // Error mask

    cin >> a >> ch;
    switch (ch)
    {
      case '<': case '>': op = ch; ch = 0; cin >> ch;
                          if (op != ch) op = ch = '?';
    }

    if (!cin.eof() && ch != '=')
    {
      op = ch; ch = 0; cin >> b >> ch;
    }

    if (!cin.eof() && ch != '=')
    {
      md = ch; ch = 0; cin >> c >> ch;
    }

    if (md == 0)
    switch (op)
    {
      case '+':   break;
      case '-':   break;
      case '*':   break;
      case '/':   break;
      case '%':   break;
      case '\\':  break;
      case '>':   break;
      case '<':   break;
      case '|':   break;
      case '&':   break;
      case '^':   break;
      case 'V':
      case 'v': if (a == 2)
                  break;
      case 0:     break;
      default:    error = 1;
    }
    else if (md == '*')
    switch (op)
    {
      case '+':   break;
      default:    error = 1;
    }
    else if (md == '\\')
    switch (op)
    {
      case '+':   break;
      default:    error = 1;
    }
    else if (md == '%')
    switch (op)
    {
      case '*':   break;
      case '\\':  break;
      default:    error = 1;
    }
    else error = 2;

    while (!cin.eof() && ch != '=')
    {
      error |= 0x4000; cin >> ch;               // Skip until =
    }

    while (!cin.eof() && (cin.get (ch), ch != '\n'))
    {
      error |= (isspace (ch) - 1) & 0x8000;     // Skip rest of line
    }

    if (error & 1)
      cout << "Error: 1st operator is either missed or unknown" << endl;

    if (error & 2)
      cout << "Error: 2nd operator is either missed or unknown" << endl;

    if (error & 0x4000)
      cout << "Error: excessive operators or operands before sign =" << endl;

    if (error & 0x8000)
      cout << "Error: misplaced operators or operands after sign =" << endl;

    if (error)
    {
      retcode = 1;
      cout << endl;
      continue;
    }

    // Generating random numbers.

    if (*pszGenerate != 0)
    {
      switch (op)
      {
        case 'V':
        case 'v':   cout << a;
                    break;
        default:    retcode |= GenerateRandom (a, radix, maxexp);
      }
      if (op != 0)
      {
        cout << endl << (char) op;
        switch (op)
        {
          case '<':
          case '>': cout << (char) op << endl << b;
                    break;
          case '\\':if (md == 0)
                    {
                      cout << endl << b;
                      break;
                    }
          default:  cout << endl;
                    retcode |= GenerateRandom (b, radix, maxexp);
        }
        if (md != 0)
        {
          cout << endl << (char) md << endl;
          retcode |= GenerateRandom (c, radix, maxexp);
        }
      }
      cout << endl << '=' << endl;
      continue;
    }
    cBigNumber cmp = a;

    long timer1 = GetTimer (0);
    total1 += timer1 - timer0;
    cerr << "Computing ";

    if (md == 0)
    switch (op)
    {
      case '+': cerr << "a+b";          a += b;         break;
      case '-': cerr << "a-b";          a -= b;         break;
      case '*': cerr << "a*b";          a *= b;         break;
      case '/': cerr << "a/b,a%b";      a /= b;         break;
      case '%': cerr << "a%b";          a %= b;         break;
      case '\\':cerr << "a\\b (pow)";   a.pow (b);      break;
      case '>': cerr << "a>>b";         a >>= b;        break;
      case '<': cerr << "a<<b";         a <<= b;        break;
      case '|': cerr << "a|b (or)";     a |= b;         break;
      case '&': cerr << "a&b (and)";    a &= b;         break;
      case '^': cerr << "a^b (xor)";    a ^= b;         break;
      case 'V':
      case 'v': cerr << "sqrt(b),b-sqrt(b)*sqrt(b)";
                                        a.setsqrt (b);  break;
    }
    else if (md == '*')
    switch (op)
    {
      case '+': cerr << "a+b*c";        a.addmul (b,c); break;
    }
    else if (md == '\\')
    switch (op)
    {
      case '+': cerr << "a+b\\c (pow)"; a += b.pow(c);  break;
    }
    else if (md == '%')
    switch (op)
    {
      case '*': cerr << "a*b%c";        a %= c; b %= c;
                                        a *= b; a %= c; break;
      case '\\':cerr << "a\\b%c (pow)"; a.powmod (b,c); break;
    }
    cerr << endl;

    long timer2 = GetTimer (0);
    total2 += timer2 - timer1;

    cout << a << endl;
    switch (op)
    {
      case '/':   cout << (c = cBigNumber::lastdivmod()) << endl;
                  break;
      case 'V':
      case 'v':   cout << (c = cBigNumber::lastrootrm()) << endl;
                  break;
    }

    long timer3 = GetTimer (0);
    total3 += timer3 - timer2;

    if (*pszCheck != 0 && md == 0
         && op != 0 && op != '|' && op != '&' && op != '>'
         && (op != '*'  || b != 0)
         && (op != '\\' || b == 2))
    {
      cerr << "Computing ";

      switch (op)
      {
        case '+': cerr << "a+b-b";        a -= b;         break;
        case '-': cerr << "a-b+b";        a += b;         break;
        case '*': cerr << "a*b/b";        a /= b;         break;
        case '/': cerr << "a/b*b+a%b";    a *= b; a += c; break;
        case '%': cerr << "a%b+a/b*b";    a += (cmp/b)*b; break;
        case '\\':cerr << "sqrt(a*a)";    if (cmp < 0) cmp.neg();
                                          a.sqrt();       break;                                   
        case '<': cerr << "(a<<b)>>b";    a >>= b;        break;
        case '^': cerr << "a^b^b";        a ^= b;         break;
        case 'V':
        case 'v': cerr << "sqrt(b)*sqrt(b)+(b-sqrt(b)*sqrt(b))";
                                   cmp=b; a *= a; a += c; break;
      }
      cerr << endl;

      long timer4 = GetTimer (0);
      total4 += timer4 - timer3;

      cout << a << endl;
      switch (op)
      {
        case '*':   cout << (c = cBigNumber::lastdivmod()) << endl;
                    break;
        case '\\':  cout << (c = cBigNumber::lastrootrm()) << endl;
                    break;
        default:    c = 0;
      }

      long timer5 = GetTimer (0);
      total5 += timer5 - timer4;
      cerr << "Comparing... ";

      if (cmp == a && c == 0)
      {
        cerr << "OK";
      }
      else
      {
        cerr << "FAIL!\b";
        retcode = 1;
      }
      cerr << endl;
    }
    cout << endl;

    if (!cout)
    {
       cerr << "Write error" << endl;
       return 1;
    }
  }

  cout << dec;

  if (*pszGenerate == 0)
  {
    long total = 0;
    cout << "Reading    " << setw (6) << total1 << " mls\n"; total += total1;
    cout << "Computing  " << setw (6) << total2 << " mls\n"; total += total2;
    cout << "Writing    " << setw (6) << total3 << " mls\n"; total += total3;
    if (*pszCheck != 0)
    {
    cout << "Checking(c)" << setw (6) << total4 << " mls\n"; total += total4;
    cout << "Checking(w)" << setw (6) << total5 << " mls\n"; total += total5;
    }
    cout << "Total      " << setw (6) << total  << " mls\n";
  }

  if (!cout)
  {
     cerr << "Write error" << endl;
     return 1;
  }

  return retcode;
}

int GenerateRandom (cBigNumber& a, size_t radix, size_t maxexp)
{
  assert (radix > 0);
  int nSign = 0;
  if (a < 0) { nSign = -1; a.neg(); }
  if (a.length() > 1)
  {
    cerr << "Can't generate random number of length greater then "
         << LONG_MAX << endl;
    cout << '0';
    return 1;
  }

  long n = a.tolong();
  cerr << "Generating random number of length " << n;
  if (maxexp != 0)
  {
    maxexp = (int) (dRand3() * maxexp);
    cerr << " and exponent " << maxexp;
  }
  cerr << endl;

  if (n != 0)
  {
    if (nSign < 0) cout << '-';
    cout << (int) (1 + (dRand3() * (radix - 1)));
    long n = a.tolong();
    while (--n != 0)
    {
      cout << (int) (dRand3() * radix);
    }
    for (; maxexp > 0; --maxexp)
    {
      cout << 0;
    }
  }
  else
  {
    cout << 0;
  }
  return 0;
}
