/* --------------------------------------------------------------
    Signed integers with unlimited range (version 1.1a).

    http://www.imach.uran.ru/cbignum

    Copyright 1999-2005 by Raul N.Shakirov, IMach of RAS(UB).
    All Rights Reserved.

    Permission has been granted to copy, distribute and modify
    software in any context without fee, including a commercial
    application, provided that the aforesaid copyright statement
    is present here as well as exhaustive description of changes.

    THE SOFTWARE IS DISTRIBUTED "AS IS". NO WARRANTY OF ANY KIND
    IS EXPRESSED OR IMPLIED. YOU USE AT YOUR OWN RISK. THE AUTHOR
    WILL NOT BE LIABLE FOR DATA LOSS, DAMAGES, LOSS OF PROFITS OR
    ANY OTHER KIND OF LOSS WHILE USING OR MISUSING THIS SOFTWARE.
-------------------------------------------------------------- */
#ifdef  _MSC_VER
#pragma warning(disable:4995)
#endif
#if __TURBOC__ > 0x410
#pragma option -xd-
#endif
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <iomanip.h>
#include <limits.h>
#include "Cbignum.h"
#define BITS            (CHAR_BIT * sizeof (long))

static  cBigNumber      a_;
static  cBigNumber      b_;
static  cBigNumber      num_;
cBigNumber cBigNumberLastDivMod;
cBigNumber cBigNumberLastRootRm;
exblockstack<long> cBigTemp::stack;

inline size_t size_max (size_t n1, size_t n2)
{
  return ((n1 > n2)? n1 : n2);
}
inline size_t size_min (size_t n1, size_t n2)
{
  return ((n1 < n2)? n1 : n2);
}

cBigNumber& cBigNumber::fit()
{
  checkindex (cBigNumberFit (EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::tab()
{
  checkexpand (exmuladd (BITS, length(), BITS * 2 + 1));
  cBigNumberTab (EXPTRTYPE(*this));
  return *this;
}

cBigNumber& cBigNumber::gc()
{
  checkshrink (length());
  return *this;
}

cBigNumber& cBigNumber::pack()
{
  reallocate (length() + (length()==0) + 1);
  return *this;
}

cBigNumber& cBigNumber::operator ++()
{
  checkexpand (length() + 1);
  checkindex (cBigNumberMInc (EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::operator --()
{
  checkexpand (length() + 1);
  checkindex (cBigNumberMDec (EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::mul2()
{
  checkexpand (length() + 1);
  checkindex (cBigNumberMMul2 (EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::div2()
{
  checkindex (cBigNumberMDiv2 (EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::pow2()
{
  checkexpand (length() * 2 + 2);
  cBigNumberMul (EXPTRTYPE(*this), EXPTRTYPE(*this), EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::sqrt()
{
  swap (cBigNumberLastRootRm);
  checkexpand (cBigNumberLastRootRm.length()/2 + 1);
  if (cBigNumberLastRootRm.hiword() >= 0)
    cBigNumberMRmSqrt (EXPTRTYPE(cBigNumberLastRootRm), EXPTRTYPE(*this));
  else
    *this = 0;
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setxor (const cBigNumber& a, const cBigNumber& b)
{
  checkindex ((a.length() >= b.length())?
    (checkexpand (a.length()),
     cBigNumberXor (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this))):
    (checkexpand (b.length()),
     cBigNumberXor (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this))));
  return *this;
}

cBigNumber& cBigNumber::setxor (const cBigNumber& a, long b)
{
  checkexpand (a.length());
  checkindex (cBigNumberXor (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setxor (long a, const cBigNumber& b)
{
  checkexpand (b.length());
  checkindex (cBigNumberXor (EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setxor (long a, long b)
{
  *this = a ^ b;
  return *this;
}

cBigNumber& cBigNumber::setand (const cBigNumber& a, const cBigNumber& b)
{
  checkindex ((a.length() >= b.length())?
    (checkexpand (a.length()),
     cBigNumberAnd (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this))):
    (checkexpand (b.length()),
     cBigNumberAnd (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this))));
  return *this;
}

cBigNumber& cBigNumber::setand (const cBigNumber& a, long b)
{
  checkexpand (a.length());
  checkindex (cBigNumberAnd (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setand (long a, const cBigNumber& b)
{
  checkexpand (b.length());
  checkindex (cBigNumberAnd (EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setand (long a, long b)
{
  *this = a & b;
  return *this;
}

cBigNumber& cBigNumber::setor  (const cBigNumber& a, const cBigNumber& b)
{
  checkindex ((a.length() >= b.length())?
    (checkexpand (a.length()),
     cBigNumberOr (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this))):
    (checkexpand (b.length()),
     cBigNumberOr (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this))));
  return *this;
}

cBigNumber& cBigNumber::setor  (const cBigNumber& a, long b)
{
  checkexpand (a.length());
  checkindex (cBigNumberOr (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                            EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setor  (long a, const cBigNumber& b)
{
  checkexpand (b.length());
  checkindex (cBigNumberOr (EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)),
                            EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setor  (long a, long b)
{
  *this = a | b;
  return *this;
}

cBigNumber& cBigNumber::setadd (const cBigNumber& a, const cBigNumber& b)
{
  checkindex ((a.length() >= b.length())?
    (checkexpand (a.length() + 1),
     cBigNumberAdd (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this))):
    (checkexpand (b.length() + 1),
     cBigNumberAdd (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this))));
  return *this;
}

cBigNumber& cBigNumber::setadd (const cBigNumber& a, long b)
{
  checkexpand (a.length() + 1);
  checkindex (cBigNumberAdd (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setadd (long a, const cBigNumber& b)
{
  checkexpand (b.length() + 1);
  checkindex (cBigNumberAdd (EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setadd (long a, long b)
{
  checkexpand (2);
  checkindex (cBigNumberAdd (EXPTRTYPE(_cBigLong(a)), EXPTRTYPE(_cBigLong(b)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setsub (const cBigNumber& a, const cBigNumber& b)
{
  checkindex ((a.length() >= b.length())?
    (checkexpand (a.length() + 1),
     cBigNumberSub  (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this))):
    (checkexpand (b.length() + 1),
     cBigNumberSubS (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this))));
  return *this;
}

cBigNumber& cBigNumber::setsub (const cBigNumber& a, long b)
{
  checkexpand (a.length() + 1);
  checkindex (cBigNumberSub (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setsub (long a, const cBigNumber& b)
{
  checkexpand (b.length() + 1);
  checkindex (cBigNumberSubS(EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setsub (long a, long b)
{
  checkexpand (2);
  checkindex (cBigNumberSub (EXPTRTYPE(_cBigLong(a)), EXPTRTYPE(_cBigLong(b)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setmul (const cBigNumber& a, const cBigNumber& b)
{
  size_t la = a.length(), lb = b.length();
  checkexpand (la + lb + 2);
  if (la >= lb)
        cBigNumberMul (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this));
  else  cBigNumberMul (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setmul (const cBigNumber& a, long b)
{
  checkexpand (a.length() + 3);
  cBigNumberMul (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)), EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setmul (long a, const cBigNumber& b)
{
  checkexpand (b.length() + 3);
  cBigNumberMul (EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)), EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setmul (long a, long b)
{
  checkexpand (4);
  cBigNumberMul (EXPTRTYPE(_cBigLong(a)), EXPTRTYPE(_cBigLong(b)),
                 EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdiv (const cBigNumber& a, const cBigNumber& b)
{
  checkexpand (a.length() + 2);
  cBigNumberLastDivMod = a;
  cBigNumberMModDiv (EXPTRTYPE(cBigNumberLastDivMod), EXPTRTYPE(b),
                     EXPTRTYPE(*this));
  cBigNumberLastDivMod.checkindex (cBigNumberLastDivMod.length());
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdiv (const cBigNumber& a, long b)
{
  checkexpand (a.length() + 2);
  cBigNumberLastDivMod = a;
  cBigNumberMModDiv (EXPTRTYPE(cBigNumberLastDivMod), EXPTRTYPE(_cBigLong(b)),
                     EXPTRTYPE(*this));
  cBigNumberLastDivMod.checkindex (cBigNumberLastDivMod.length());
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdiv (long a, const cBigNumber& b)
{
  checkexpand (3);
  cBigNumberLastDivMod = a;
  cBigNumberMModDiv (EXPTRTYPE(cBigNumberLastDivMod), EXPTRTYPE(b),
                     EXPTRTYPE(*this));
  cBigNumberLastDivMod.checkindex (cBigNumberLastDivMod.length());
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdiv (long a, long b)
{
  checkexpand (3);
  cBigNumberLastDivMod = a;
  cBigNumberMModDiv (EXPTRTYPE(cBigNumberLastDivMod), EXPTRTYPE(_cBigLong(b)),
                     EXPTRTYPE(*this));
  cBigNumberLastDivMod.checkindex (cBigNumberLastDivMod.length());
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setmod (const cBigNumber& a, const cBigNumber& b)
{
  *this = a; *this %= b;
  return *this;
}

cBigNumber& cBigNumber::setmod (const cBigNumber& a, long b)
{
  *this = a; *this %= b;
  return *this;
}

cBigNumber& cBigNumber::setmod (long a, const cBigNumber& b)
{
  *this = a; *this %= b;
  return *this;
}

cBigNumber& cBigNumber::setmod (long a, long b)
{
  *this = a % b;
  return *this;
}

cBigNumber& cBigNumber::operator <<= (const cBigNumber& b)
{
  if (b.hiword() < 0) return (*this >>= (-b));
  num_.setdivmod (b_ = b, BITS);
  size_t n = (size_t)b_.loword();
  checkexpand ((num_.length() <= 1 &&
             (unsigned long)num_.loword() < UINT_MAX-length())?
            length() + (size_t)num_.loword() + 2: UINT_MAX);
  for (; n > 0; --n)  cBigNumberMMul2 (EXPTRTYPE(*this));
  cBigNumberCopyShl (EXPTRTYPE(*this), (size_t)num_.loword(),
                     EXPTRTYPE(*this));
  checkindex (length());
  EXDEBUG (num_ = 0);
  return *this;
}

cBigNumber& cBigNumber::operator <<= (long b)
{
  if (b < 0) return (*this >>= (-b));
  size_t n = (size_t) b % BITS; b /= BITS;
  checkexpand ((unsigned long)b < UINT_MAX-length()?
            length() + (size_t)b + 2: UINT_MAX);
  for (; n > 0; --n) cBigNumberMMul2 (EXPTRTYPE(*this));
  cBigNumberCopyShl (EXPTRTYPE(*this), (size_t)b, EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setshl (const cBigNumber& a, const cBigNumber& b)
{
  *this = a; return *this <<= b;
}

cBigNumber& cBigNumber::setshl (const cBigNumber& a, long b)
{
  *this = a; return *this <<= b;
}

cBigNumber& cBigNumber::setshl (long a, const cBigNumber& b)
{
  *this = a; return *this <<= b;
}

cBigNumber& cBigNumber::setshl (long a, long b)
{
  *this = a; return *this <<= b;
}

cBigNumber& cBigNumber::operator >>= (const cBigNumber& b)
{
  if (b.hiword() < 0) return (*this <<= (-b));
  num_.setdivmod (b_ = b, BITS);
  size_t n = (size_t)b_.loword();
  for (; n > 0; --n)  cBigNumberMDiv2 (EXPTRTYPE(*this));
  cBigNumberCopyShr (EXPTRTYPE(*this),
        ((num_.length() <= 1 && (unsigned long)num_.loword() <= UINT_MAX)?
         (size_t)num_.loword(): UINT_MAX),
        EXPTRTYPE(*this));
  checkindex (length());
  EXDEBUG (num_ = 0);
  return *this;
}

cBigNumber& cBigNumber::operator >>= (long b)
{
  if (b < 0) return (*this <<= (-b));
  size_t n = (size_t) b % BITS; b /= BITS;
  for (; n > 0; --n) cBigNumberMDiv2 (EXPTRTYPE(*this));
  cBigNumberCopyShr (EXPTRTYPE(*this),
        (((unsigned long)b <= UINT_MAX)? (size_t)b: UINT_MAX),
                     EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setshr (const cBigNumber& a, const cBigNumber& b)
{
  *this = a; return *this >>= b;
}

cBigNumber& cBigNumber::setshr (const cBigNumber& a, long b)
{
  *this = a; return *this >>= b;
}

cBigNumber& cBigNumber::setshr (long a, const cBigNumber& b)
{
  *this = a; return *this >>= b;
}

cBigNumber& cBigNumber::setshr (long a, long b)
{
  *this = a; return *this >>= b;
}

cBigNumber& cBigNumber::setpow_()
{
  if (b_.hiword() < 0)
  {
    a_.setdiv (1,a_); b_.neg();
  }
  num_.setmul (b_, a_.bits());
  size_t n = (
        (num_.length() <= 1 && (unsigned long)num_.loword() <= UINT_MAX)?
        (size_t)(num_.loword() / BITS + 3): UINT_MAX);
  checkexpand (n); a_.checkexpand (n);
  cBigNumberPow (EXPTRTYPE(a_), EXPTRTYPE(b_), EXPTRTYPE(*this));
  checkindex (length());
  EXDEBUG (a_ = 0);
  EXDEBUG (b_ = 0);
  EXDEBUG (num_ = 0);
  return *this;
}

cBigNumber& cBigNumber::setpow (const cBigNumber& a, const cBigNumber& b)
{
  a_ = a; b_ = b; return setpow_();
}

cBigNumber& cBigNumber::setpow (const cBigNumber& a, long b)
{
  a_ = a; b_ = b; return setpow_();
}

cBigNumber& cBigNumber::setpow (long a, const cBigNumber& b)
{
  a_ = a; b_ = b; return setpow_();
}

cBigNumber& cBigNumber::setpow (long a, long b)
{
  a_ = a; b_ = b; return setpow_();
}

cBigNumber& cBigNumber::setpowmod_ (const cBigNumber &mod)
{
  if (b_.hiword() < 0)
  {
    a_.setdiv (1,a_); b_.neg();
  }
  size_t n = mod.length() * 2 + 1;
  checkexpand (n); a_.checkexpand (n);
  cBigNumberPowMod (EXPTRTYPE(a_), EXPTRTYPE(b_), EXPTRTYPE(mod),
                    EXPTRTYPE(*this));
  checkindex (length());
  EXDEBUG (a_ = 0);
  EXDEBUG (b_ = 0);
  return *this;
}

cBigNumber& cBigNumber::setpowmod (const cBigNumber& a, const cBigNumber& b,
                                   const cBigNumber& mod)
{
  a_.setmod (a, mod); b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setpowmod (const cBigNumber& a, long b,
                                   const cBigNumber& mod)
{
  a_.setmod (a, mod); b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setpowmod (long a, const cBigNumber& b,
                                   const cBigNumber& mod)
{
  a_.setmod (a, mod); b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setpowmod (long a, long b,
                                   const cBigNumber& mod)
{
  a_.setmod (a, mod); b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setpowmod_ (long mod)
{
  if (b_.hiword() < 0)
  {
    a_.setdiv (1,a_); b_.neg();
  }
  checkexpand (3); a_.checkexpand (3);
  cBigNumberPowMod (EXPTRTYPE(a_), EXPTRTYPE(b_), EXPTRTYPE(_cBigLong(mod)),
                    EXPTRTYPE(*this));
  checkindex (length());
  EXDEBUG (a_ = 0);
  EXDEBUG (b_ = 0);
  return *this;
}

cBigNumber& cBigNumber::setpowmod (const cBigNumber& a, const cBigNumber& b,
                                   long mod)
{
  a_.setmod (a, mod); b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setpowmod (const cBigNumber& a, long b,
                                   long mod)
{
  a_.setmod (a, mod); b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setpowmod (long a, const cBigNumber& b,
                                   long mod)
{
  a_.setmod (a, mod); b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setpowmod (long a, long b,
                                   long mod)
{
  a_.setmod (a, mod); b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setsqrt (const cBigNumber& a)
{
  checkexpand (a.length()/2 + 1);
  cBigNumberLastRootRm = a;
  if (a.hiword() >= 0)
    cBigNumberMRmSqrt (EXPTRTYPE(cBigNumberLastRootRm), EXPTRTYPE(*this));
  else
    *this = 0;
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setsqrt (long a)
{
  checkexpand (1);
  cBigNumberLastRootRm = a;
  if (a >= 0)
    cBigNumberMRmSqrt (EXPTRTYPE(cBigNumberLastRootRm), EXPTRTYPE(*this));
  else
    *this = 0;
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::addmul (const cBigNumber& a, const cBigNumber& b)
{
  size_t la = a.length(),lb = b.length();
  checkexpand (size_max (la + lb + 2, length() + 1));
  if (la >= lb)
        cBigNumberMAddMul (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this));
  else  cBigNumberMAddMul (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::addmul (const cBigNumber& a, long b)
{
  checkexpand (size_max (a.length() + 3, length() + 1));
  cBigNumberMAddMul (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                     EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::addmul (long a, const cBigNumber& b)
{
  checkexpand (size_max (b.length() + 3, length() + 1));
  cBigNumberMAddMul (EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)),
                     EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::addmul (long a, long b)
{
  checkexpand (size_max (4, length() + 1));
  cBigNumberMAddMul (EXPTRTYPE(_cBigLong(a)), EXPTRTYPE(_cBigLong(b)),
                     EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdivmod (cBigNumber& a, const cBigNumber& b)
{
  checkexpand (a.length() + 2);
  cBigNumberMModDiv (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this));
  a.checkindex (a.length());
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdivmod (cBigNumber& a, long b)
{
  checkexpand (a.length() + 2);
  cBigNumberMModDiv (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                     EXPTRTYPE(*this));
  a.checkindex (a.length());
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setsqrtrm (cBigNumber& a)
{
  checkexpand (a.length()/2 + 1);
  if (a.hiword() >= 0)
    cBigNumberMRmSqrt (EXPTRTYPE(a), EXPTRTYPE(*this));
  else
    *this = 0;
  a.checkindex (a.length());
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::addmultab (const cBigNumber& a, const cBigNumber& b,
                                   size_t k)
{
  checkexpand (size_max (a.length() + b.length() + k + 1, length() + 1));
  assert (b.hiword() >= 0);
  cBigNumberMAddMulTab (EXPTRTYPE(a), EXPTRTYPE(b), k, EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::addmultab (const cBigNumber& a, long b, size_t k)
{
  checkexpand (size_max (a.length() + k + 2, length() + 1));
  assert (b >= 0);
  cBigNumberMAddMulTab (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)), k,
                        EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdivtab (const cBigNumber& a, const cBigNumber& b)
{
  checkexpand (a.length() + 2);
  cBigNumberLastDivMod = a;
  cBigNumberMModDivTab (EXPTRTYPE(cBigNumberLastDivMod), EXPTRTYPE(b),
                        EXPTRTYPE(*this));
  cBigNumberLastDivMod.checkindex (cBigNumberLastDivMod.length());
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdivtab (long a, const cBigNumber& b)
{
  checkexpand (3);
  cBigNumberLastDivMod = a;
  cBigNumberMModDivTab (EXPTRTYPE(cBigNumberLastDivMod), EXPTRTYPE(b),
                        EXPTRTYPE(*this));
  cBigNumberLastDivMod.checkindex (cBigNumberLastDivMod.length());
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setmodtab (const cBigNumber& a, const cBigNumber& b)
{
  *this = a;
  cBigNumberMModTab (EXPTRTYPE(*this), EXPTRTYPE(b));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setmodtab (long a, const cBigNumber& b)
{
  *this = a;
  cBigNumberMModTab (EXPTRTYPE(*this), EXPTRTYPE(b));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdivmodtab (cBigNumber& a, const cBigNumber& b)
{
  checkexpand (a.length() + 2);
  cBigNumberMModDivTab (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this));
  a.checkindex (a.length());
  checkindex (length());
  return *this;
}

int     b_SPRP  (
        const cBigNumber& n,
        const cBigNumber& b
)
{
  if (n < 2 || (n.loword() & 1) == 0)
    return 0;

  static cBigNumber t;
  static cBigNumber nMinus1;
  static cBigNumber a;
  t = nMinus1 = n - 1; a = 0;

  while( !(t.loword() & 1) ) { t >>= 1; ++a; }

  cBigNumber test; test.setpowmod (b, t, n);

  if (test == 1 || test == nMinus1) return 2;

  if (a)
  {
    while (--a)
    {
      test.pow2(); test %= n; if (test == nMinus1) return 2;
    }
  }

  return 0;
}
