/* --------------------------------------------------------------
    Signed integers with unlimited range (version 1.1a).
    Service functions in machine-independent form.

    http://www.imach.uran.ru/cbignum

    Copyright 1999-2005 by Raul N.Shakirov, IMach of RAS(UB).
    All Rights Reserved.

    Permission has been granted to copy, distribute and modify
    software in any context without fee, including a commercial
    application, provided that the aforesaid copyright statement
    is present here as well as exhaustive description of changes.

    THE SOFTWARE IS DISTRIBUTED "AS IS". NO WARRANTY OF ANY KIND
    IS EXPRESSED OR IMPLIED. YOU USE AT YOUR OWN RISK. THE AUTHOR
    WILL NOT BE LIABLE FOR DATA LOSS, DAMAGES, LOSS OF PROFITS OR
    ANY OTHER KIND OF LOSS WHILE USING OR MISUSING THIS SOFTWARE.
-------------------------------------------------------------- */
#ifdef  _MSC_VER
#pragma warning(disable:4995)
#endif
#include <ctype.h>
#include <stdlib.h>
#include <iomanip.h>
#include <limits.h>
#include "Cbignum.h"
#include <assert.h>
#define BITS            (CHAR_BIT * sizeof (long))

#define SHIFT_TAB_MIN   (3)
#define SHIFT_TAB_OPT   ((sizeof(size_t))==2? 15-3: 480-3)
#define SHIFT_TAB_HIGH  (SHIFT_TAB_OPT + SHIFT_TAB_OPT / 2)
#define SHIFT_TAB_MAX   (EXSIZE_T_MAX / (1024 * sizeof (long)))

inline size_t size_max (size_t n1, size_t n2)
{
  return ((n1 > n2)? n1 : n2);
}
inline size_t size_min (size_t n1, size_t n2)
{
  return ((n1 < n2)? n1 : n2);
}

#ifdef  __cplusplus
extern "C" {
#endif

void    cBigNumberDump (
                const   EXPTR(long) p1,
                ostream& stream
        )
{
  char fill  = stream.fill  ('0');
  long flags = stream.flags (ios::hex);

  size_t n1 = (size_t)(*p1);
  for (; n1; n1--) stream <<" "<< setw(8) << (unsigned long) p1 [n1];
  stream << endl;

  stream.fill  (fill);
  stream.flags (flags);
}

void    cBigNumberERange (
                const   EXPTR(long) p1
)
{
  cerr <<"cBigNumber value is out of range:";
  cBigNumberDump (p1, cerr);
  abort();
}

int     cBigNumberIsFit (
                const   EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);
  if (n1)
  {
    if (n1 > 1) return ((p1 [n1] != (p1 [n1-1] >> (BITS-1))));
  }
  return (n1);
}

int     cBigNumberBits (
                const   EXPTR(long) p1
        )
{
  size_t n1 = (int)(*p1);
  if (n1)
  {
    long sign = p1 [n1];
    do
    {
      unsigned long num = p1 [n1];
      if (sign < 0) num = ~num;
      if (num)
      {
        if (n1 <= (INT_MAX / BITS))
        {
          n1 *= BITS;
          if (num < (1L<<(BITS/2)))
          {
            n1 -= BITS; do ++n1; while ((num >>= 1) > 0);
          }
          else if ((long)num > 0)
          {
            do --n1; while ((long)(num <<= 1) > 0);
          }
        }
        else n1 = INT_MAX;
        break;
      }
    }
    while (--n1 != 0);
  }
  return (n1);
}

int     cLongBits (long num)
{
  int n1 = BITS;
  if (num < 0) num = ~num;
  if (num >= (1L<<(BITS/2)))
  {
    do --n1; while ((num <<= 1) > 0);
  }
  else
  {
    n1 -= BITS; if (num) do ++n1; while ((num >>= 1) > 0);
  }
  return (n1);
}

size_t  cBigNumberExpWords (
                const   EXPTR(long) p1
)
{
  size_t n1 = (size_t)(*p1);
  size_t n = 0;
  for (; n != n1; n++)
  {
    if (p1 [n + 1] == 0) break;
  }
  return (n);
}

int     cBigNumberComp  (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2
        )
{
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);
  register long lp1 = 0;
  register long lp2 = 0;

  if (n1) lp1 = p1 [n1--];
  if (n2) lp2 = p2 [n2--];

  if (n1 == n2)
  {
    if (lp1 != lp2) goto NotSameSignWord;
  }
  else if (n1 > n2)
  {
    lp2 >>= (BITS-1); ++n2;
    if (lp1 != lp2) goto NotSameSignWord;
    if (n1 > n2)
    {
      do
      {
        lp1 = p1 [n1]; --n1;
        if (lp1 != lp2) goto NotSame;
      }
      while (n1 > n2);
    }
  }
  else
  {
    lp1 >>= (BITS-1); ++n1;
    if (lp1 != lp2) goto NotSameSignWord;
    if (n1 < n2)
    {
      do
      {
        lp2 = p2 [n2]; --n2;
        if (lp1 != lp2) goto NotSame;
      }
      while (n1 < n2);
    }
  }

  if (n1)
  {
    do
    {
      lp1 = p1 [n1]; lp2 = p2 [n1]; --n1;
      if (lp1 != lp2) goto NotSame;
    }
    while (n1);
  }

  return (0);
NotSame:
  return (((unsigned long)lp1 >= (unsigned long)lp2) * 2 - 1);
NotSameSignWord:
  return ((lp1 >= lp2) * 2 - 1);
}

int     cBigNumberCompHigh (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2
        )
{
  assert (cBigNumberIsFit(p1));
  assert (cBigNumberIsFit(p2));
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);
  assert (n1 >= n2);
  assert (n2 > 0);

  register long lp1 = p1 [n1--];
  register long lp2 = p2 [n2--];
  if (lp1 != lp2) return ((lp1 >= lp2) * 2 - 1);

  if (n2)
  do
  {
    lp1 = p1 [n1--]; lp2 = p2 [n2--];
    if (lp1 != lp2)
      return (((unsigned long)lp1 >= (unsigned long)lp2) * 2 - 1);
  }
  while (n2);

  if (n1)
  do
  {
    if (p1 [n1--] != 0) return 1;
  }
  while (n1);

  return (0);
}

inline  long    cBigNumberNot0  (
                const EXPTR(long) p1
                )
{
  assert (cBigNumberIsFit(p1));
  size_t n1 = (size_t)(*p1);
  return (p1 [n1] | (long)(n1 >> 1));
}


void    cBigNumberCopy (
                const   EXPTR(long) p1,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p++ = *p1++);
  if (n1)
  do
    { *p++ = *p1++; }
  while (--n1 != 0);
}

void    cBigNumberCompl (
                const   EXPTR(long) p1,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p++ = *p1++);
  if (n1)
  do
    { *p++ = ~*p1++; }
  while (--n1 != 0);
}

void    cBigNumberCopyShl (
                const   EXPTR(long) p1,
                        size_t      k1,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1);
  if (n1 <= 1 && p1 [n1] == 0) k1 = 0;
  p1 += n1; *p = n1 + k1; p += (n1 + k1);
  if (n1)
  do
    { *p-- = *p1--; }
  while (--n1 != 0);
  if (k1)
  do
    { *p-- = 0; }
  while (--k1 != 0);
}

void    cBigNumberCopyShr (
                const   EXPTR(long) p1,
                        size_t      k1,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1);
  if (n1 > k1)
  {
    p1 += (k1 + 1);
    *p++ = (n1 -= k1);
    do
      { *p++ = *p1++; }
    while (--n1 != 0);
  }
  else { *p++ = 1; *p = p1 [n1] >> (BITS-1); }
}

void    cBigNumberCutOut (
                const   EXPTR(long) p1,
                        size_t      k1,
                        size_t      n,
                        EXPTR(long) p
        )
{
  assert ((size_t)(*p1) >= k1 + n);
  *p++ = n + 1;
  if (n > 0)
  {
    p1 += (k1 + 1);
    do
      { *p++ = *p1++; }
    while (--n != 0);
  }
  *p = 0;
}

size_t  _cBigNumberNeg (
                const   EXPTR(long) p1,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  p++;
  size_t n = n1;
  if (n)
  do
  {
    long lp;
    if ((*p = lp = -*p1) != 0)
    {
      if (--n != 0)
      {
        do
          { ++p1; ++p; *p = lp = ~*p1; }
        while (--n != 0);
        return (n1 - (lp == (p[-1]>>(BITS-1))));
      }
      if (lp != LONG_MIN) return (n1);
      p [1] = 0; return (n1 + 1);
    }
  }
  while ((++p1, ++p, --n) != 0);

  return (n1);
}

size_t  _cBigNumberMInc (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n = n1;

  if (n)
  {
    do
    {
      long lp;
      if ((*p1 = lp = *p1 + 1) != 0)
      {
        if (--n != 0)
        {
          if (--n != 0) return (n1);
          return (n1 - (p1[1]==(lp>>(BITS-1))));
        }
        if (lp != LONG_MIN) return (n1);
        p1 [1] = 0; return (n1 + 1);
      }
    }
    while ((++p1, --n) != 0);
  }
  else { *p1 = 1; n1 = 1; }

  return (n1);
}

size_t  _cBigNumberMDec (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n = n1;

  if (n)
  {
    do
    {
      long lp;
      if ((*p1 = lp = *p1 - 1) != ~0)
      {
        if (--n != 0)
        {
          if (--n != 0) return (n1);
          return (n1 - (p1[1]==(lp>>(BITS-1))));
        }
        if (lp != LONG_MAX) return (n1);
        p1 [1] = ~0; return (n1 + 1);
      }
    }
    while ((++p1, --n) != 0);
  }
  else { *p1 = -1; n1 = 1; }

  return (n1);
}

size_t  _cBigNumberMMul2 (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n = n1;
  long num = 0;
  if (n)
  do
  {
    long pass = (unsigned long)num >> (BITS-1);
    num = *p1;
    *p1++ = (num << 1) | pass;
  }
  while (--n != 0);

  num >>= (BITS-2);
  n1 += (num != (*p1 = num >> 1));

  return (n1);
}

size_t  _cBigNumberMul2 (
                const   EXPTR(long) p1,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  p++;
  long num = 0;
  size_t n = n1;

  if (n)
  do
  {
    long pass = (unsigned long)num >> (BITS-1);
    num = *p1++;
    *p++ = (num << 1) | pass;
  }
  while (--n != 0);

  num >>= (BITS-2);
  n1 += (num != (*p = num >> 1));

  return (n1);
}

size_t  _cBigNumberMDiv2 (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);
  size_t n = n1;
  p1 += n;
  if (n)
  {
    if (--n != 0)
    {
      long num = *p1;
      n1 -= (num == (*p1-- = (num >> 1)));
      do
      {
        long pass = num << (BITS-1);
        num = *p1;
        *p1-- = ((unsigned long)num>>1) | pass;
      }
      while (--n != 0);
    }
    else
    {
      (*p1) >>= 1;
    }
  }
  return (n1);
}

size_t  _cBigNumberDiv2 (
                const   EXPTR(long) p1,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1);
  size_t n = n1;
  p1 += n;
  p  += n;
  if (n)
  {
    if (--n != 0)
    {
      long num = *p1--;
      n1 -= (num == (*p-- = (num >> 1)));
      do
      {
        long pass = num << (BITS-1);
        num = *p1--;
        *p-- = ((unsigned long)num>>1) | pass;
      }
      while (--n != 0);
    }
    else
    {
      *p = (*p1) >> 1;
    }
  }
  return (n1);
}

size_t  _cBigNumberFit  (
                        EXPTR(long) p1
        )
{
  register size_t n1 = (size_t)(*p1);
  p1 += n1;
  register long pass;
  register long lp;

  if (n1)
  {
    lp = *p1;
    do continue;
    while (--n1 != 0 &&
        (pass = lp, lp = p1 [-1], --p1, pass == (lp>>(BITS-1))));
  }
  else { p1 [1] = 0; }

  return (n1 + 1);
}

size_t  _cBigNumberFitTo (
                        EXPTR(long) p1,
                        size_t n
        )
{
  register size_t n1 = (size_t)(*p1);
  register long pass = p1 [n1] >> (BITS-1);

  do
  {
     p1 [++n1] = pass;
  }
  while (n1 < n);

  return (n1);
}

void    cBigNumberTab  (
                        EXPTR(long) p
        )
{
  size_t n = cBigNumberFit (p);
  int ns = BITS - 1;
  p [n + 1] = p [n] >> (BITS-1);
  n += 2;
  EXPTR(long) pp = p;
  do
  {
    pp += n;
    cBigNumberMul2 (p, pp);
    p = pp;
  }
  while (--ns > 0);
}

size_t  _cBigNumberXor  (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n2 = (size_t)(*p2++);
  p++;
  assert (n1 >= n2);

  {
    register long lp2 = 0;

    {
      size_t n;
      if ((n = n2) != 0)
      {
        do { *p++ = *p1++ ^ (lp2 = *p2++); }
        while (--n != 0);
      }
    }

    {
      lp2 >>= (BITS-1);
      size_t n;
      if ((n = n1 - n2) != 0)
      {
        do { *p++ = *p1++ ^ lp2; }
        while (--n != 0);
      }
    }
    --p;
  }

  {
    register long pass;
    register long lp;
    register size_t n = n1;
    if (n)
    {
      lp = *p;
      do continue;
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, pass == (lp>>(BITS-1))));
    }
    else { p [1] = 0; }

    return (n + 1);
  }
}

size_t  _cBigNumberAnd  (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n2 = (size_t)(*p2++);
  p++;
  assert (n1 >= n2);

  {
    register long lp2 = 0;

    {
      size_t n;
      if ((n = n2) != 0)
      {
        do { *p++ = *p1++ & (lp2 = *p2++); }
        while (--n != 0);
      }
    }

    {
      lp2 >>= (BITS-1);
      size_t n;
      if ((n = n1 - n2) != 0)
      {
        do { *p++ = *p1++ & lp2; }
        while (--n != 0);
      }
    }
    --p;
  }

  {
    register long pass;
    register long lp;
    register size_t n = n1;
    if (n)
    {
      lp = *p;
      do continue;
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, pass == (lp>>(BITS-1))));
    }
    else { p [1] = 0; }

    return (n + 1);
  }
}

size_t  _cBigNumberOr  (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n2 = (size_t)(*p2++);
  p++;
  assert (n1 >= n2);

  {
    register long lp2 = 0;

    {
      size_t n;
      if ((n = n2) != 0)
      {
        do { *p++ = *p1++ | (lp2 = *p2++); }
        while (--n != 0);
      }
    }

    {
      lp2 >>= (BITS-1);
      size_t n;
      if ((n = n1 - n2) != 0)
      {
        do { *p++ = *p1++ | lp2; }
        while (--n != 0);
      }
    }
    --p;
  }

  {
    register long pass;
    register long lp;
    register size_t n = n1;
    if (n)
    {
      lp = *p;
      do continue;
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, pass == (lp>>(BITS-1))));
    }
    else { p [1] = 0; }

    return (n + 1);
  }
}

size_t  _cBigNumberAdd  (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n2 = (size_t)(*p2++);
  p++;
  assert (n1 >= n2);

  register unsigned long pass = 0;
  {
    register long lp1 = 0;
    register long lp2 = 0;

    {
      size_t n;
      if ((n = n2) != 0)
      {
        do
        {
          lp1 = *p1++; lp2 = *p2++;
          pass = (~(*p++ = (pass>>(BITS-1)) + lp1 + lp2) & (lp1 ^ lp2))
                                                         | (lp1 & lp2);
        }
        while (--n != 0);
      }
    }

    {
      size_t n;
      if ((n = n1 - n2) != 0)
      {
        register const long *pp = p1;
        if ((lp2 & LONG_MIN) == 0)
        {
          do
          {
            lp1 = *pp++;
            pass = ~(*p++ = (pass>>(BITS-1)) + lp1) & lp1;
          }
          while (--n != 0);
        }
        else
        {
          do
          {
            lp1 = *pp++;
            pass = ~(*p++ = (pass>>(BITS-1)) + lp1 - 1) | lp1;
          }
          while (--n != 0);
        }
      }
    }
    --p;

    pass = (pass>>(BITS-1)) + ((lp1)>>(BITS-1)) + ((lp2)>>(BITS-1));
  }

  {
    register long lp = *p;
    register size_t n = n1;
    if (n && (long)pass == (lp>>(BITS-1)))
    {
      do continue;
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, (long)pass == (lp>>(BITS-1))));
    }
    else { p [1] = pass; }

    return (n + 1);
  }
}

size_t  _cBigNumberSub  (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n2 = (size_t)(*p2++);
  p++;
  assert (n1 >= n2);

  register unsigned long pass = ~0UL;
  {
    register long lp1 = 0;
    register long lp2 = ~0;

    {
      size_t n;
      if ((n = n2) != 0)
      {
        do
        {
          lp1 = *p1++; lp2 = ~(*p2++);
          pass = (~(*p++ = (pass>>(BITS-1)) + lp1 + lp2) & (lp1 ^ lp2))
                                                         | (lp1 & lp2);
        }
        while (--n != 0);
      }
    }

    {
      size_t n;
      if ((n = n1 - n2) != 0)
      {
        register const long *pp = p1;
        if ((lp2 & LONG_MIN) == 0)
        {
          do
          {
            lp1 = *pp++;
            pass = ~(*p++ = (pass>>(BITS-1)) + lp1) & lp1;
          }
          while (--n != 0);
        }
        else
        {
          do
          {
            lp1 = *pp++;
            pass = ~(*p++ = (pass>>(BITS-1)) + lp1 - 1) | lp1;
          }
          while (--n != 0);
        }
      }
    }
    --p;

    pass = (pass>>(BITS-1)) + ((lp1)>>(BITS-1)) + ((lp2)>>(BITS-1));
  }

  {
    register long lp = *p;
    register size_t n = n1;
    if (n && (long)pass == (lp>>(BITS-1)))
    {
      do continue;
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, (long)pass == (lp>>(BITS-1))));
    }
    else { p [1] = pass; }

    return (n + 1);
  }
}

size_t  _cBigNumberSubS (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n2 = (size_t)(*p2++);
  p++;
  assert (n1 >= n2);

  register unsigned long pass = ~0UL;
  {
    register long lp1 = ~0;
    register long lp2 = 0;

    {
      size_t n;
      if ((n = n2) != 0)
      {
        do
        {
          lp1 = ~(*p1++); lp2 = *p2++;
          pass = (~(*p++ = (pass>>(BITS-1)) + lp1 + lp2) & (lp1 ^ lp2))
                                                         | (lp1 & lp2);
        }
        while (--n != 0);
      }
    }

    {
      size_t n;
      if ((n = n1 - n2) != 0)
      {
        register const long *pp = p1;
        if ((lp2 & LONG_MIN) == 0)
        {
          do
          {
            lp1 = ~(*pp++);
            pass = ~(*p++ = (pass>>(BITS-1)) + lp1) & lp1;
          }
          while (--n != 0);
        }
        else
        {
          do
          {
            lp1 = ~(*pp++);
            pass = ~(*p++ = (pass>>(BITS-1)) + lp1 - 1) | lp1;
          }
          while (--n != 0);
        }
      }
    }
    --p;

    pass = (pass>>(BITS-1)) + ((lp1)>>(BITS-1)) + ((lp2)>>(BITS-1));
  }

  {
    register long lp = *p;
    register size_t n = n1;
    if (n && (long)pass == (lp>>(BITS-1)))
    {
      do continue;
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, (long)pass == (lp>>(BITS-1))));
    }
    else { p [1] = pass; }

    return (n + 1);
  }
}

static  size_t  _cBigNumberSkip = 0;

size_t  cBigNumberSkipLow0 (
                const   EXPTR(long) p1
)
{
  size_t n1 = (size_t)(*p1);
  size_t n = 1;

  if (n1 != 0)
  do
  {
    if (p1 [n] != 0) break;
    n++;
  }
  while (--n1 != 0);

  return _cBigNumberSkip = n;
}

#ifndef _CBIGNUM_ASM

void    cBigNumberMAddM (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        size_t k2
        )
{
  const EXPTR(long) pp = p1 + (size_t)(*p1);
  size_t n2 = (size_t)(*p2);
  assert (_cBigNumberSkip > 0);
  assert ((size_t)(*p1) >= n2 + k2);
  assert (n2 >= _cBigNumberSkip);
  p1 += k2;

  register unsigned long pass = 0;
  {
    const EXPTR(long) pp = p1 + n2;
    p1 += _cBigNumberSkip;
    p2 += _cBigNumberSkip;
    register long lp1;
    register long lp2;

    for (;;)
    {
      lp1 = *p1;        lp2 = *p2;
      pass += lp1;      lp1 ^= lp2;
      pass += lp2;      lp1 = ~lp1;
      *p1 = pass;       pass |= lp1;
      lp1 &= lp2;       pass = ~pass;
      ++p2;             pass |= lp1;
      if (p1 >= pp)     break;
      ++p1;             pass >>= (BITS-1);
    }
    if ((long)(lp2 ^= pass) >= 0) return;
  }

  {
    register long lp1;
    if ((long)pass < 0)
    {
      do
      {
        if (p1 >= pp) break;
        lp1 = *++p1;
      }
      while ((~(*p1 = lp1 + 1) & lp1) < 0);
    }
    else
    {
      do
      {
        if (p1 >= pp) break;
        lp1 = *++p1;
      }
      while ((~(*p1 = lp1 - 1) | lp1) >= 0);
    }
  }
}

size_t  _cBigNumberMSubD (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        size_t k2
        )
{
  size_t n2 = (size_t)(*p2);
  assert (_cBigNumberSkip > 0);
  assert ((p1 [(size_t)(*p1)] ^ p2 [n2]) >= 0);
  assert ((size_t)(*p1) >= n2 + k2);
  assert (n2 >= _cBigNumberSkip);
  p1 += k2;

  register long pass = 0;
  {
    const EXPTR(long) pp = p1 + n2;
    p1 += _cBigNumberSkip;
    p2 += _cBigNumberSkip;
    register long lp1;
    register long lp2;

    for (;;)
    {
      lp1 = *p1;        lp2 = *p2;
      pass += lp1;      lp1 ^= lp2;
      pass -= lp2;      lp2 = ~lp2;
      *p1 = pass;       lp2 &= lp1;
      if (p1 >= pp)     break;
      lp2 = ~lp2;       pass |= lp1;
      ++p1;             pass &= lp2;
      ++p2;             pass >>= (BITS-1);
    }
  }

  {
    register size_t n1 = n2 + k2;

    for (;;)
    {
      register long lp1;
      register long lp2;
      if (--n1 == 0)    break;
      lp2 = p1 [-1];    --p1;
      lp1 = pass;       pass = lp2;
      lp2 >>= (BITS-1);
      if (lp1 != lp2)   break;
    }
    return (n1 + 1);
  }
}

size_t  _cBigNumberMMul2M (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);
  assert (_cBigNumberSkip > 0);
  assert (n1 >= _cBigNumberSkip);

  const EXPTR(long) pp = p1 + n1;
  p1 += _cBigNumberSkip;
  long num = 0;
  do
  {
    long pass = (unsigned long)num >> (BITS-1);
    num = *p1;
    *p1++ = (num << 1) | pass;
  }
  while (p1 <= pp);

  num >>= (BITS-2);
  n1 += (num != (*p1 = num >> 1));

  return (n1);
}

size_t  _cBigNumberMDiv2D (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);
  assert (_cBigNumberSkip > 0);
  assert (n1 >= _cBigNumberSkip);
  assert ((p1 [_cBigNumberSkip] & 1L) == 0);

  const EXPTR(long) pp = p1 + _cBigNumberSkip;
  p1 += n1;
  long num = *p1;
  n1 -= (n1 > 1) & (num == (*p1-- = num >> 1));
  while (p1 >= pp)
  {
    long pass = num << (BITS-1);
    num = *p1;
    *p1-- = ((unsigned long)num >> 1) | pass;
  }

  return (n1);
}

#endif

void    cBigNumberMShrM (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);
  assert (n1 > 1);
  *p1 = --n1;
  do
    { p1 [1] = p1 [2]; p1++; }
  while ((int)--n1 > 0);
}

void    cBigNumberMShlD (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);
  assert (n1 > 0);
  *p1  = n1 + 1;
  do
    { p1 [n1 + 1] = p1 [n1]; }
  while ((int)--n1 > 0);
  p1 [1] = 0;
}

static  cBigNumber   cBigBuf1;
static  cBigNumber   cBigBuf2;
static  cBigNumber   cBigBuf3;
static  cBigNumber   cBigBuf4;

void    cBigNumberMAddMulShl (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        size_t      k,
                        EXPTR(long) p
        )
{
  assert (_cBigNumberSkip == 0);
  assert (cBigNumberIsFit (p1));
  assert (p1 != p);
  assert (p2 != p);
  assert (p2 != p1);
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);

  if (n1 >= cBigNumberSkipLow0 (p1) && n2 != 0)
  {
    cBigNumberFitTo (p, n1 + n2 + k + 1);
    --n2;
    size_t k2 = 0;
    for (; k2 != n2; k2++)
    {
      long num = p2 [k2 + 1];
      if (num == 0) continue;
      long mask = 1;
      for (;; cBigNumberMMul2M (p1))
      {
        if (num & mask) cBigNumberMAddM (p, p1, k + k2);
        if ((mask <<= 1) == 0) break;
      }
      cBigNumberMMul2M (p1);
      cBigNumberMShrM (p1);
    }
    {
      long num = p2 [k2 + 1];
      long mask = (num != 0);
      if (num > 0)
      do
        mask <<= 1;
      while ((num <<= 1) > 0);
      for (;; cBigNumberMMul2M (p1))
      {
        if (num & mask) cBigNumberMAddM (p, p1, k + k2);
        if ((mask <<= 1) == 0) break;
      }
    }
  }
  cBigNumberFit (p);
  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberMAddMulTab (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        size_t      k,
                        EXPTR(long) p
        )
{
  assert (_cBigNumberSkip == 0);
  assert (cBigNumberIsFit (p1));
  assert (p1 != p);
  assert (p2 != p);
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);

  if (n1 >= cBigNumberSkipLow0 (p1))
  {
    cBigNumberFitTo (p, n1 + n2 + k + 1);
    n1 += 2;
    for (size_t k2 = 0; k2 != n2; k2++)
    {
      long num = p2 [k2 + 1];
      long mask = 1;
      const EXPTR(long) pp1 = p1;
      for (;; pp1 += n1)
      {
        if (num & mask) cBigNumberMAddM (p, pp1, k + k2);
        if ((mask <<= 1) == 0) break;
      }
    }
  }

  cBigNumberFit (p);
  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberMAddMul (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  assert (_cBigNumberSkip == 0);
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);

  cBigBuf1.checkexpand ((n2 <= SHIFT_TAB_MIN)? (n1 + 3):
                        (n1 < SHIFT_TAB_HIGH)? ((n1 + 3) * BITS + 1):
                  size_max (n1 + 3, (SHIFT_TAB_HIGH + 3) * BITS + 1));
  cBigBuf2.checkexpand (n2 + 1);
  EXPTR(long) pp1 = EXPTRTYPE(cBigBuf1);
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);

  if (p2 [n2] >= 0)
  {
    cBigNumberCopy (p1, pp1);
    cBigNumberCopy (p2, pp2);
  }
  else
  {
    cBigNumberNeg  (p1, pp1);
    cBigNumberNeg  (p2, pp2);
    assert (pp2 [(size_t)(*pp2)] >= 0);
    *pp2 = n2;
  }

  if (n2 > SHIFT_TAB_MIN)
  {
    size_t k = 0;
    if (k + (SHIFT_TAB_HIGH) < n1)
    {
      cBigBuf3.checkexpand ((SHIFT_TAB_OPT + 3) * BITS + 1);
      EXPTR(long) pp3 = EXPTRTYPE(cBigBuf3);
      do
      {
        cBigNumberCutOut (pp1, k, SHIFT_TAB_OPT, pp3);
        cBigNumberTab (pp3);
        cBigNumberMAddMulTab (pp3, pp2, k, p);
        k += SHIFT_TAB_OPT;
      }
      while (k + (SHIFT_TAB_HIGH) < n1);
      cBigNumberCopyShr (pp1, k, pp1);
    }
    cBigNumberTab (pp1);
    cBigNumberMAddMulTab (pp1, pp2, k, p);
    EXDEBUG (cBigBuf1 = 0);
    EXDEBUG (cBigBuf2 = 0);
    EXDEBUG (cBigBuf3 = 0);
    EXDEBUG (_cBigNumberSkip = 0);
    return;
  }

  if (n1 >= cBigNumberSkipLow0 (pp1) && n2 != 0)
  {
    cBigNumberFitTo (p, n1 + n2 + 1);
    --n2;
    size_t k2 = 0;
    for (; k2 != n2; k2++)
    {
      long num = pp2 [k2 + 1];
      if (num == 0) continue;
      long mask = 1;
      for (;; cBigNumberMMul2M (pp1))
      {
        if (num & mask) cBigNumberMAddM (p, pp1, k2);
        if ((mask <<= 1) == 0) break;
      }
      cBigNumberMMul2M (pp1);
      cBigNumberMShrM (pp1);
    }
    {
      long num = pp2 [k2 + 1];
      long mask = (num != 0);
      if (num > 0)
      do
        mask <<= 1;
      while ((num <<= 1) > 0);
      for (;; cBigNumberMMul2M (pp1))
      {
        if (num & mask) cBigNumberMAddM (p, pp1, k2);
        if ((mask <<= 1) == 0) break;
      }
    }
  }
  cBigNumberFit (p);
  EXDEBUG (cBigBuf1 = 0);
  EXDEBUG (cBigBuf2 = 0);
  EXDEBUG (cBigBuf3 = 0);
  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberMul (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  assert (_cBigNumberSkip == 0);
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);

  cBigBuf1.checkexpand ((n2 <= SHIFT_TAB_MIN)? (n1 + 3):
                        (n1 < SHIFT_TAB_HIGH)? ((n1 + 3) * BITS + 1):
                  size_max (n1 + 3, (SHIFT_TAB_HIGH + 3) * BITS + 1));
  cBigBuf2.checkexpand (n2 + 1);
  EXPTR(long) pp1 = EXPTRTYPE(cBigBuf1);
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);

  if (p2 [n2] >= 0)
  {
    cBigNumberCopy (p1, pp1);
    cBigNumberCopy (p2, pp2);
  }
  else
  {
    cBigNumberNeg  (p1, pp1);
    cBigNumberNeg  (p2, pp2);
    assert (pp2 [(size_t)(*pp2)] >= 0);
    *pp2 = n2;
  }

  p [0] = 0;

  if (n2 > SHIFT_TAB_MIN)
  {
    size_t k = 0;
    if (k + (SHIFT_TAB_HIGH) < n1)
    {
      cBigBuf3.checkexpand ((SHIFT_TAB_OPT + 3) * BITS + 1);
      EXPTR(long) pp3 = EXPTRTYPE(cBigBuf3);
      do
      {
        cBigNumberCutOut (pp1, k, SHIFT_TAB_OPT, pp3);
        cBigNumberTab (pp3);
        cBigNumberMAddMulTab (pp3, pp2, k, p);
        k += SHIFT_TAB_OPT;
      }
      while (k + (SHIFT_TAB_HIGH) < n1);
      cBigNumberCopyShr (pp1, k, pp1);
    }
    cBigNumberTab (pp1);
    cBigNumberMAddMulTab (pp1, pp2, k, p);
    EXDEBUG (cBigBuf1 = 0);
    EXDEBUG (cBigBuf2 = 0);
    EXDEBUG (cBigBuf3 = 0);
    EXDEBUG (_cBigNumberSkip = 0);
    return;
  }

  if (n1 >= cBigNumberSkipLow0 (pp1) && n2 != 0)
  {
    cBigNumberFitTo (p, n1 + n2 + 1);
    --n2;
    size_t k2 = 0;
    for (; k2 != n2; k2++)
    {
      long num = pp2 [k2 + 1];
      if (num == 0) continue;
      long mask = 1;
      for (;; cBigNumberMMul2M (pp1))
      {
        if (num & mask) cBigNumberMAddM (p, pp1, k2);
        if ((mask <<= 1) == 0) break;
      }
      cBigNumberMMul2M (pp1);
      cBigNumberMShrM (pp1);
    }
    {
      long num = pp2 [k2 + 1];
      long mask = (num != 0);
      if (num > 0)
      do
        mask <<= 1;
      while ((num <<= 1) > 0);
      for (;; cBigNumberMMul2M (pp1))
      {
        if (num & mask) cBigNumberMAddM (p, pp1, k2);
        if ((mask <<= 1) == 0) break;
      }
    }
  }
  cBigNumberFit (p);
  EXDEBUG (cBigBuf1 = 0);
  EXDEBUG (cBigBuf2 = 0);
  EXDEBUG (cBigBuf3 = 0);
  EXDEBUG (_cBigNumberSkip = 0);
}

size_t  cBigNumberMaskDiv0;
void    cBigNumberDiv0 (void)
{
  cBigNumberMaskDiv0 /= cBigNumberMaskDiv0;
}

void    cBigNumberMModDivTab (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  assert (_cBigNumberSkip == 0);
  assert (cBigNumberIsFit (p1));
  assert (cBigNumberIsFit (p2));
  assert (p1 != p);
  assert (p2 != p);
  assert (p2 != p1);
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);
  size_t n = ((n1 > n2)? n1 - n2: 0) + 1;

  *p = n + 1; p += n; p [1] = 0;

  cBigNumberSkipLow0 (p2);
  if (cBigNumberNot0 (p2))
  {
    assert ((p1 [n1] ^ p2 [n2]) >= 0);
    long lt = (p1 [n1] < 0) * 2 - 1;
    const EXPTR(long) pp2 = p2;
    n2 += 2;
    do
    {
      --n;
      pp2 += n2 * BITS;
      unsigned long mask = (unsigned long)
                           (1L << (BITS-1));
      *p = 0;
      do
      {
        pp2 -= n2;
        assert (*pp2 >= (long)(n2 - 2) && *pp2 <  (long) n2);
        int diff = n - (size_t)(*p1)
                     + (size_t)(*pp2);
        if (diff <= 0 && (diff < 0 ||
            cBigNumberCompHigh (p1, pp2) != lt))
        {
          cBigNumberMSubD (p1, pp2, n);
          (*p) += mask;
        }
      }
      while ((mask >>= 1) != 0);
      --p;
    }
    while (n);
    assert (cBigNumberComp (p1, pp2) == lt);
  }
  else
  {
    cBigNumberDiv0();
    do { *p-- = 0; } while (--n);
  }
  cBigNumberFit (p);
  assert (p [(size_t)(*p)] >= 0);
  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberMModDiv (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  assert (_cBigNumberSkip == 0);
  assert (p1 != p);

  size_t n1 = (size_t) cBigNumberFit (p1);
  size_t n2 = (size_t)(*p2);
  int nt = (int) ((SHIFT_TAB_MIN + n2 - n1) & (n1 - SHIFT_TAB_MAX));

  cBigBuf2.checkexpand ((nt < 0)? ((n2 + 3) * BITS + 1): (n2 + 3));
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);

  long signdiv;
  if ((p1 [n1] ^ p2 [n2]) >= 0)
    { cBigNumberCopy (p2, pp2); signdiv = 1; }
  else
    { cBigNumberNeg (p2, pp2); signdiv = -1; }

  if (nt < 0)
  {
    cBigNumberTab (pp2);
    cBigNumberMModDivTab (p1, pp2, p);
    if (signdiv < 0) cBigNumberNeg (p, p);
    EXDEBUG (cBigBuf2 = 0);
    EXDEBUG (_cBigNumberSkip = 0);
    return;
  }

  n2 = cBigNumberFit (pp2);
  size_t n = ((n1 > n2)? n1 - n2: 0) + 1;

  *p = n + 1; p += n; p [1] = 0;

  cBigNumberSkipLow0 (pp2);
  if (cBigNumberNot0 (pp2))
  {
    assert ((p1 [n1] ^ pp2 [n2]) >= 0);
    long lt = (p1 [n1] < 0) * 2 - 1;
    do
    {
      --n;
      *p = 0;
      cBigNumberMShlD (pp2);
      unsigned long mask = (1UL << (BITS-1));
      do
      {
        int diff = cBigNumberMDiv2D (pp2) + n -
                   (size_t)(*p1);
        if (diff <= 0 && (diff < 0 ||
            cBigNumberCompHigh (p1, pp2) != lt))
        {
          cBigNumberMSubD (p1, pp2, n);
          (*p) += mask;
        }
      }
      while ((mask >>= 1) != 0);
      --p;
    }
    while (n);
    assert (cBigNumberComp (p1, pp2) == lt);
  }
  else
  {
    cBigNumberDiv0();
    do { *p-- = 0; } while (--n);
  }
  cBigNumberFit (p);
  assert (p [(size_t)(*p)] >= 0);
  if (signdiv < 0) cBigNumberNeg (p, p);
  EXDEBUG (cBigBuf2 = 0);
  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberMModTab (
                        EXPTR(long) p1,
                const   EXPTR(long) p2
        )
{
  assert (_cBigNumberSkip == 0);
  assert (cBigNumberIsFit (p1));
  assert (cBigNumberIsFit (p2));
  assert (p1 != p2);
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);
  size_t n = ((n1 > n2)? n1 - n2: 0) + 1;

  cBigNumberSkipLow0 (p2);
  if (cBigNumberNot0 (p2))
  {
    assert ((p1 [n1] ^ p2 [n2]) >= 0);
    long lt = (p1 [n1] < 0) * 2 - 1;
    const EXPTR(long) pp2 = p2;
    n2 += 2;
    do
    {
      --n;
      pp2 += n2 * BITS;
      unsigned long mask = (1UL << (BITS-1));
      do
      {
        pp2 -= n2;
        assert (*pp2 >= (long)(n2 - 2) && *pp2 <  (long) n2);
        int diff = n - (size_t)(*p1)
                     + (size_t)(*pp2);
        if (diff <= 0 && (diff < 0 ||
            cBigNumberCompHigh (p1, pp2) != lt))
        {
          cBigNumberMSubD (p1, pp2, n);
        }
      }
      while ((mask >>= 1) != 0);
    }
    while (n);
    assert (cBigNumberComp (p1, pp2) == lt);
  }
  else
  {
    cBigNumberDiv0();
  }
  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberMMod (
                        EXPTR(long) p1,
                const   EXPTR(long) p2
        )
{
  assert (_cBigNumberSkip == 0);

  size_t n1 = (size_t) cBigNumberFit (p1);
  size_t n2 = (size_t)(*p2);
  int nt = (int) ((SHIFT_TAB_MIN + n2 - n1) & (n1 - SHIFT_TAB_MAX));

  cBigBuf2.checkexpand ((nt < 0)? ((n2 + 3) * BITS + 1): (n2 + 3));
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);

  if ((p1 [n1] ^ p2 [n2]) >= 0)
    { cBigNumberCopy (p2, pp2); }
  else
    { cBigNumberNeg  (p2, pp2); }

  if (nt < 0)
  {
    cBigNumberTab (pp2);
    cBigNumberMModTab (p1, pp2);
    EXDEBUG (cBigBuf2 = 0);
    EXDEBUG (_cBigNumberSkip = 0);
    return;
  }

  n2 = cBigNumberFit (pp2);
  size_t n = ((n1 > n2)? n1 - n2: 0) + 1;
  cBigNumberSkipLow0 (pp2);
  if (cBigNumberNot0 (pp2))
  {
    assert ((p1 [n1] ^ pp2 [n2]) >= 0);
    long lt = (p1 [n1] < 0) * 2 - 1;
    do
    {
      --n;
      cBigNumberMShlD (pp2);
      unsigned long mask = (1UL << (BITS-1));
      do
      {
        int diff = cBigNumberMDiv2D (pp2) + n -
                   (size_t)(*p1);
        if (diff <= 0 && (diff < 0 ||
            cBigNumberCompHigh (p1, pp2) != lt))
        {
          cBigNumberMSubD (p1, pp2, n);
        }
      }
      while ((mask >>= 1) != 0);
    }
    while (n);
    assert (cBigNumberComp (p1, pp2) == lt);
  }
  else
  {
    cBigNumberDiv0();
  }
  EXDEBUG (cBigBuf2 = 0);
  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberPow (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  assert (p1 != p);
  assert (p2 != p);
  assert (p2 != p1);
  size_t n2 = (size_t)(*p2);
  assert (p2 [n2] >= 0);
  p2++;

  p [0] = 1; p [1] = 1;

  if (n2 != 0)
  {
    for (;;)
    {
      long num = *p2++;
      long mask = 1;
      if (--n2 == 0)
      {
        if (num == 0) break;
        do
          mask <<= 1;
        while ((num <<= 1) > 0);
      }
      for (;;)
      {
        if (num & mask)
          cBigNumberMul (p1, p, p);
        if ((mask <<= 1) == 0) break;
        cBigNumberMul (p1, p1, p1);
      }
      if (n2 == 0) break;
      cBigNumberMul (p1, p1, p1);
    }
  }
}

void    cBigNumberPowMod (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                const   EXPTR(long) mod,
                        EXPTR(long) p
        )
{
  assert (p1 != p);
  assert (p2 != p);
  assert (p2 != p1);
  size_t n2 = (size_t)(*p2);
  assert (p2 [n2] >= 0);
  p2++;
  size_t nmod = (size_t)(*mod);

  cBigBuf4.checkexpand (exmuladd (BITS, nmod, BITS * 3 + 1));
  EXPTR(long) pmod = EXPTRTYPE(cBigBuf4);

  if (mod [nmod] >= 0)
    { cBigNumberCopy (mod, pmod); }
  else
    { cBigNumberNeg  (mod, pmod); }

  cBigNumberTab (pmod);
  if (cBigNumberNot0 (pmod) == 0) n2 = 0;

  p [0] = 1; p [1] = 1;
  cBigNumberMModTab (p, pmod);

  if (n2 != 0 && p [1] != 0)
  {

    long sign = p1 [(size_t)(*p1)];
    if (sign < 0)
    {
      cBigNumberNeg (p1, p1);
      assert (p1 [(size_t)(*p1)] >= 0);
      sign = (*p2 << (BITS-1));
    }
    cBigNumberMModTab (p1, pmod);

    for (;;)
    {
      long num = *p2++;
      long mask = 1;
      if (--n2 == 0)
      {
        if (num == 0) break;
        do
          mask <<= 1;
        while ((num <<= 1) > 0);
      }
      for (;;)
      {
        if (num & mask)
        {
          if (*p1 > *p) cBigNumberMul (p1, p, p);
          else          cBigNumberMul (p, p1, p);
          cBigNumberMModTab (p, pmod);
        }
        if ((mask <<= 1) == 0) break;
        cBigNumberMul (p1, p1, p1);
        cBigNumberMModTab (p1, pmod);
      }
      if (n2 == 0) break;
      cBigNumberMul (p1, p1, p1);
      cBigNumberMModTab (p1, pmod);
    }
    if (sign < 0) cBigNumberNeg (p, p);
  }

  EXDEBUG (cBigBuf4 = 0);
}

void    cBigNumberMRmSqrt (
                        EXPTR(long) p1,
                        EXPTR(long) p
        )
{
  assert (_cBigNumberSkip == 0);
  assert (p1 != p);

  size_t n1 = (size_t) cBigNumberFit (p1);
  assert (p1 [n1] >= 0);

  cBigBuf2.checkexpand (n1 + 1);
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);
  *pp2 = 0;
  cBigNumberFitTo (pp2, n1);

  int nbit = (cBigNumberBits (p1) + 1) & ~1;
  if (nbit < 0) cBigNumberERange (p1);
  if ((nbit -= 2) >= 0)
  {

    {
      size_t nw = nbit / BITS;
      long lw = (1L<<(nbit % BITS));
      pp2 [nw + 1] += lw;
      assert (pp2 [(size_t)(*pp2)] >= 0);
      p1  [nw + 1] -= lw;
      assert (p1  [(size_t)(*p1)]  >= 0);
    }

    while ((nbit -= 2) >= 0)
    {
      size_t nw = nbit / BITS;
      long lw = (1L<<(nbit % BITS));
      pp2 [nw + 1] += lw;
      _cBigNumberSkip = nw + 1;
      if (cBigNumberComp (p1, pp2) < 0)
      {
        pp2 [nw + 1] -= lw;
        cBigNumberMDiv2D (pp2);
      }
      else
      {
        cBigNumberMSubD (p1, pp2, 0);
        assert (p1 [(size_t)(*p1)] >= 0);
        assert (cBigNumberIsFit (p1));
        pp2 [nw + 1] -= lw;
        cBigNumberMDiv2D (pp2);
        pp2 [nw + 1] += lw;
      }
      assert (pp2 [(size_t)(*pp2)] >= 0);
      assert (cBigNumberIsFit (pp2));
    }
  }
  cBigNumberFit (pp2);
  cBigNumberCopy (pp2, p);

  assert (cBigNumberComp (
    (_cBigNumberSkip = 1, cBigNumberMMul2M (pp2), pp2), p1) >= 0);

  EXDEBUG (_cBigNumberSkip = 0);
}

#ifdef  __cplusplus
}
#endif
