/* --------------------------------------------------------------
    Timer and priority functions.

    Copyright 1998-2003 by Dr. Raul N.Shakirov, IMach of RAS(UB).
    All Rights Reserved.

    Permission has been granted to copy, distribute and modify
    software in any context without fee, including a commercial
    application, provided that the aforesaid copyright statement
    is present here as well as exhaustive description of changes.

    THE SOFTWARE IS DISTRIBUTED "AS IS". NO WARRANTY OF ANY KIND
    IS EXPRESSED OR IMPLIED. YOU USE AT YOUR OWN RISK. THE AUTHOR
    WILL NOT BE LIABLE FOR DATA LOSS, DAMAGES, LOSS OF PROFITS OR
    ANY OTHER KIND OF LOSS WHILE USING OR MISUSING THIS SOFTWARE.
-------------------------------------------------------------- */
#include <time.h>       /* Header file for standard time functions */
#include "Gettimer.h"   /* Header file for GetTimer function */

#ifdef  __cplusplus
extern "C" {
#endif/*__cplusplus*/

/* --------------------------------------------------------------
    Function GetTimer reads system clock and returns difference
    between current time and provided time value in mls.

    Parameter:
    timer           Time in mls to subtract from current time.

    Returns:        Interval in mls between current time and
                    given time value.
-------------------------------------------------------------- */

long    GetTimer (long timer)
{
  /*  Get time  */

  long time = ((long)(clock() * (1000. / CLOCKS_PER_SEC)) - timer);

  /*  Correct time after midnight  */

  if (time < 0) time += (3600 * 24 * 1000L);

  return time;
}

/* --------------------------------------------------------------
    Function SetLowPriority sets low priority class for the
    application. At first start, is also stores existing
    priority class, which is set for application by system.

    Function ResetPriority either resets the stored priority
    class or does noting if no priority class is stored.

    NOTE:   Implemented for Win32 if macro __STDC__ is not set.
-------------------------------------------------------------- */

#ifndef __STDC__
#ifdef  _WIN32
#define COMPILING_FOR_WIN32
#endif/*_WIN32*/
#ifdef  __WIN32__
#define COMPILING_FOR_WIN32
#endif/*__WIN32__*/
#endif/*__STDC__*/

#ifdef  COMPILING_FOR_WIN32
#ifdef  _MSC_VER
#pragma warning(disable:4115)
#endif/*_MSC_VER*/
#include <windows.h>
DWORD   dwPriorityClass = 0;
#endif/*COMPILING_FOR_WIN32*/

void    SetHighPriority (void)
{
#ifdef  COMPILING_FOR_WIN32
  if (dwPriorityClass == 0)
  {
    dwPriorityClass = GetPriorityClass (GetCurrentProcess());
  }
  if (dwPriorityClass != 0)
  {
    SetPriorityClass (GetCurrentProcess(), HIGH_PRIORITY_CLASS);
  }
#endif/*COMPILING_FOR_WIN32*/
}

void    SetIdlePriority (void)
{
#ifdef  COMPILING_FOR_WIN32
  if (dwPriorityClass == 0)
  {
    dwPriorityClass = GetPriorityClass (GetCurrentProcess());
  }
  if (dwPriorityClass != 0)
  {
    SetPriorityClass (GetCurrentProcess(), IDLE_PRIORITY_CLASS);
  }
#endif/*COMPILING_FOR_WIN32*/
}

void    ResetPriority (void)
{
#ifdef  COMPILING_FOR_WIN32
  if (dwPriorityClass != 0)
  {
    SetPriorityClass (GetCurrentProcess(), dwPriorityClass);
  }
#endif/*COMPILING_FOR_WIN32*/
}

#ifdef  __cplusplus
}
#endif/*__cplusplus*/
